package org.ow2.petals.registry.database;

import java.io.IOException;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.HashMap;
import java.util.Map;

import javax.naming.CompositeName;
import javax.naming.NameNotFoundException;
import javax.naming.NamingException;

import org.apache.commons.dbcp.ConnectionFactory;
import org.apache.commons.dbcp.DriverManagerConnectionFactory;
import org.apache.commons.dbcp.PoolableConnectionFactory;
import org.apache.commons.dbcp.PoolingDataSource;
import org.apache.commons.pool.impl.GenericObjectPool;
import org.ow2.petals.util.LoggingUtil;

public class MySqlClient extends RegistryClient {

    public MySqlClient(String connectionUrl, String username, String password,
            int connectionPoolSize, int batchSize, boolean checkTables, LoggingUtil log)
            throws NamingException {
        super(connectionUrl, username, password, connectionPoolSize, batchSize, checkTables, log);

        this.establishConnection();

        if (checkTables)
            this.checkTables();
    }

    private void establishConnection() throws NamingException {

        final ClientKey key = new ClientKey(this.connectionUrl, this.username, this.password);
        // final ClientKey key = new ClientKey(this.username, this.password);

        synchronized (DATASOURCES) {
            this.dataSource = DATASOURCES.get(key);
            if (this.dataSource == null) {

                // Load PostgreSQL JDBC Driver
                try {
                    DriverManager.registerDriver(new com.mysql.jdbc.Driver());
                } catch (final SQLException e2) {
                    final NamingException nex = new NamingException(
                            "PostgreSQL JDBC Driver registration failed:" + e2.getMessage());
                    nex.setRootCause(e2);
                    throw nex;
                }

                // Create a pooled datasource using Apache commons-dbcp
                final GenericObjectPool connectionPool = new GenericObjectPool(null);
                connectionPool.setMaxActive(this.connectionPoolSize);

                // Factory creating simple database connections
                final ConnectionFactory connectionFactory = new DriverManagerConnectionFactory(
                        this.connectionUrl, this.username, this.password);

                // Factory wrappercreating poolable database connections. A
                // poolable connection is returned to the pool when closing
                // it.
                final PoolableConnectionFactory poolableConnectionFactory = new PoolableConnectionFactory(
                        connectionFactory, connectionPool, null, // TODO: Add a
                                                                 // prepared
                                                                 // statement
                                                                 // pooling.
                        null, // No validation when returning connecting.
                        false, // By default, the connection is read/write.
                        false); // By default, the auto-commit mode is disabled.

                // The datasource providing poolable connection.
                this.dataSource = new PoolingDataSource(connectionPool);

                DATASOURCES.put(key, this.dataSource);
            }
        }
    }

    private void checkTables() throws NamingException {

        try {
            final Connection jdbcConnection = this.dataSource.getConnection();

            try {

                final Statement stCheck = jdbcConnection.createStatement();
                boolean isStructureCreated;
                try {
                    stCheck.executeQuery("SELECT 1 FROM JNDI_CONTEXT");
                    isStructureCreated = true;
                } catch (SQLException e) {
                    isStructureCreated = false;
                } finally {
                    stCheck.close();
                }

                // A savepoint is needed here in case of an exception occurs on
                // the previous statement.
                jdbcConnection.commit();

                if (!isStructureCreated) {
                    this.log.info("Creating database structure");
                    final Statement stInit = jdbcConnection.createStatement();
                    try {
                        this.log.debug("Creating table JNDI_CONTEXT");
                        stInit
                                .executeUpdate("CREATE TABLE JNDI_CONTEXT (CONTEXT_ID SERIAL NOT NULL, CONTEXT VARCHAR(767) NOT NULL, CONSTRAINT CONTEXT_PK PRIMARY KEY (CONTEXT_ID))");
                        this.log.debug("Creating index on table JNDI_CONTEXT");
                        stInit
                                .executeUpdate("CREATE UNIQUE INDEX CTX_CTXNAME ON JNDI_CONTEXT (CONTEXT)");
                        this.log.debug("Creating table JNDI_BIND");
                        stInit
                                .executeUpdate("CREATE TABLE JNDI_BIND (CONTEXT_ID BIGINT UNSIGNED NOT NULL, NAME VARCHAR(512) NOT NULL, OBJECT BLOB  NOT NULL, CONSTRAINT BINDING_PK PRIMARY KEY (CONTEXT_ID, NAME))");
                        this.log.debug("Creating reference constraint");
                        stInit
                                .executeUpdate("ALTER TABLE JNDI_BIND ADD CONSTRAINT FK_CONTEXT FOREIGN KEY(CONTEXT_ID) REFERENCES JNDI_CONTEXT(CONTEXT_ID)");
                        this.log.debug("Creating initial context in JNDI_CONTEXT");
                        stInit.executeUpdate("INSERT INTO JNDI_CONTEXT (CONTEXT) VALUES ('"
                                + ROOT_CONTEXT + "')");
                        this.log.info("Database structure created");
                    } finally {
                        stInit.close();
                    }

                    jdbcConnection.commit();
                }
            } finally {
                jdbcConnection.close();
            }
        } catch (final SQLException e) {
            e.printStackTrace();
            final NamingException nex = new NamingException(e.getMessage());
            nex.setRootCause(e);
            throw nex;
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @seeorg.ow2.petals.registry.database.RegistryClient#list(javax.naming.
     * CompositeName)
     */
    public Map<String, Object> list(final CompositeName contextName) throws NameNotFoundException,
            NamingException {
        try {
            final String contextNameStr = contextName.toString();
            final Connection jdbcConnection = this.dataSource.getConnection();

            try {
                jdbcConnection.setAutoCommit(false);

                // 1 - We get the bound objects
                final Map<String, Object> namedObjects = this.getBoundObjects(contextNameStr,
                        jdbcConnection);
                if (namedObjects.isEmpty()) {
                    // No named object in the context. Perhaps the context
                    // exists and is empty. So we check that it exists.
                    if (!this.isContextExist(contextNameStr, jdbcConnection)) {
                        throw new NameNotFoundException(contextName + " does not exist.");
                    }
                }

                jdbcConnection.commit();

                return namedObjects;

            } finally {
                // We rollback potential updates. In case of successfully
                // processing a commit must occur before this point
                jdbcConnection.rollback();
                jdbcConnection.close();
            }
        } catch (final SQLException e) {
            throw new NamingException(e.getMessage());
        } catch (final IOException e) {
            throw new NamingException(e.getMessage());
        } catch (final ClassNotFoundException e) {
            throw new NamingException(e.getMessage());
        }
    }

    private final static String getBoundObjectsSql = new StringBuffer(
            "SELECT JNDI_BIND.NAME, JNDI_BIND.OBJECT ").append(
            "FROM JNDI_CONTEXT INNER JOIN JNDI_BIND ").append(
            "ON JNDI_CONTEXT.CONTEXT_ID = JNDI_BIND.CONTEXT_ID ").append(
            "WHERE JNDI_CONTEXT.CONTEXT = ? ").toString();

    // Caution, we use a regexp to match context according to '^contextName/...'
    private final static String getBoundSubContextSql = new StringBuffer(
            "SELECT JNDI_CONTEXT.CONTEXT ").append("FROM JNDI_CONTEXT ").append(
            "" + "WHERE JNDI_CONTEXT.CONTEXT REGEXP ? ").toString();

    private Map<String, Object> getBoundObjects(final String contextName,
            final Connection jdbcConnection) throws SQLException, IOException,
            ClassNotFoundException {

        final Map<String, Object> namedObjects = new HashMap<String, Object>(this.batchSize);

        final PreparedStatement pstGetBoundObject = jdbcConnection
                .prepareStatement(getBoundObjectsSql);
        final PreparedStatement pstGetBoundSubContext = jdbcConnection
                .prepareStatement(getBoundSubContextSql);
        try {

            // Get bound objects (that are not Context)
            {
                pstGetBoundObject.setString(1, contextName);
                pstGetBoundObject.setFetchSize(this.batchSize);
                final ResultSet rs = pstGetBoundObject.executeQuery();
                try {
                    while (rs.next()) {
                        namedObjects.put(rs.getString(1), RegistryClient.readObject(rs, 2));
                    }

                } finally {
                    rs.close();
                }
            }

            // Get bound sub-context
            {
                final String regexp = (contextName == null || contextName.length() == 0) ? "^[^/]+$"
                        : "^" + contextName + "/{1}";
                pstGetBoundSubContext.setString(1, regexp);
                pstGetBoundSubContext.setFetchSize(this.batchSize);

                final ResultSet rs = pstGetBoundSubContext.executeQuery();
                try {
                    // For performance reasons, the test about contextName is
                    // put outside the loop
                    if (contextName == null || contextName.length() == 0) {
                        while (rs.next()) {
                            // The SQL request returns the full context name of
                            // sub-contexts. So we must remove the part
                            // corresponding to the current context.
                            namedObjects.put(rs.getString(1), null);
                        }
                    } else {
                        while (rs.next()) {
                            // The SQL request returns the full context name of
                            // sub-contexts. So we must remove the part
                            // corresponding to the current context.
                            final String fullSubContextName = rs.getString(1);
                            namedObjects.put(
                                    fullSubContextName.substring(contextName.length() + 1), null);
                        }
                    }

                    return namedObjects;

                } finally {
                    rs.close();
                }
            }

        } finally {
            pstGetBoundObject.close();
            pstGetBoundSubContext.close();
        }
    }

}
