package org.ow2.petals.registry.database;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;

import javax.naming.CompositeName;
import javax.naming.ContextNotEmptyException;
import javax.naming.NameAlreadyBoundException;
import javax.naming.NameNotFoundException;
import javax.naming.NamingException;
import javax.naming.NotContextException;
import javax.sql.DataSource;

import org.ow2.petals.util.LoggingUtil;

public abstract class RegistryClient {

    protected final static String ROOT_CONTEXT = "";

    protected class ClientKey {
        private final String connectionUrl;

        private final String username;

        private final String password;

        public ClientKey(final String connectionUrl, final String username,
        // public ClientKey(final String username,

                final String password) {
            this.connectionUrl = connectionUrl;
            this.username = username;
            this.password = password;
        }

        /**
         * @see java.lang.Object#equals(java.lang.Object)
         */
        @Override
        public boolean equals(Object obj) {
            if (this == obj) {
                return true;
            } else if (obj == null || obj.getClass() != this.getClass()) {
                return false;
            } else {
                ClientKey key2 = (ClientKey) obj;
                return this.connectionUrl.equals(key2.connectionUrl)
                        && this.username.equals(key2.username)
                        && this.password.equals(key2.password);
                // return this.username.equals(key2.username) &&
                // this.password.equals(key2.password);

            }
        }

        /**
         * @see java.lang.Object#hashCode()
         */
        @Override
        public int hashCode() {
            int hash = 7;
            hash = 31 * hash + this.connectionUrl.hashCode();
            hash = 31 * hash + this.username.hashCode();
            hash = 31 * hash + this.password.hashCode();
            return hash;
        }
    }

    protected static ConcurrentMap<ClientKey, DataSource> DATASOURCES = new ConcurrentHashMap<ClientKey, DataSource>();

    protected final String connectionUrl;

    protected final String username;

    protected final String password;

    protected final int connectionPoolSize;

    protected final int batchSize;

    protected DataSource dataSource;

    protected LoggingUtil log;

    public RegistryClient(final String connectionUrl, final String username, final String password,
            final int connectionPoolSize, final int batchSize, final boolean checkTables,
            final LoggingUtil log) throws NamingException {

        this.log = log;

        this.connectionUrl = connectionUrl;
        this.username = username;
        this.password = password;
        this.connectionPoolSize = connectionPoolSize;
        this.batchSize = batchSize;

    }

    public void close() {
        final ClientKey key = new ClientKey(this.connectionUrl, this.username, this.password);
        // final ClientKey key = new ClientKey(this.username, this.password);

        DATASOURCES.remove(key);
    }

    /**
     * Bind a name to an object or create a subcontext
     * 
     * @param contextName
     *            Context name, not null.
     * @param objectName
     *            Name of the object to bind in the context
     *            <code>contextName</code>. If <code>null</code>, the subcontext
     *            <code>contextName</code> will be created.
     * @param object
     *            Object to bind. If <code>null</code>, the subcontext
     *            <code>contextName</code> will be created.
     * @throws NameAlreadyBoundException
     *             The name is already bound
     * @throws NamingException
     *             An SQL error occurs.
     */
    public void bind(final CompositeName contextName, final String objectName, final Object object)
            throws NameAlreadyBoundException, NamingException {

        try {
            final String contextNameStr = contextName.toString();
            final Connection jdbcConnection = this.dataSource.getConnection();

            try {
                jdbcConnection.setAutoCommit(false);

                if (object == null) {
                    // We must create a sub-context:
                    // 1 - We check if it exist
                    // Note: It is needed to lock the table 'context'
                    // to prevent concurrent creation of the same context.
                    if (this.isContextExist(contextNameStr, jdbcConnection)) {
                        // The subcontext exists
                        throw new NameAlreadyBoundException(contextNameStr + " already exists.");
                    }

                    // 2 - We create it
                    this.createContext(contextNameStr, jdbcConnection);
                } else {
                    // We must create a bound object
                    // 1 - We check if the bound object exist
                    // Note: It is not needed to lock the row of bound
                    // object to prevent concurrent writing because a
                    // object should be bound by an only one PEtALS
                    // instance.
                    if (this.isBoundObjectExistWithoutLock(contextNameStr, objectName,
                            jdbcConnection)) {
                        // The bound object exists
                        throw new NameAlreadyBoundException(contextNameStr + " already exists.");
                    }
                    // 2 - We create it
                    this.createBoundObject(contextNameStr, objectName, object, jdbcConnection);
                }

                jdbcConnection.commit();

            } finally {
                // We rollback potential updates. In case of successfully
                // processing a commit must occur before this point
                jdbcConnection.rollback();
                jdbcConnection.close();
            }
        } catch (final SQLException e) {
            throw new NamingException(e.getMessage());
        } catch (final IOException e) {
            throw new NamingException(e.getMessage());
        }

    }

    /**
     * Rebind a name to an object, overwriting an existing one
     * 
     * @param contextName
     *            Context name, not null.
     * @param objectName
     *            Name of the object binding, not null.
     * @param object
     *            Object to bind, not null.
     * @throws NamingException
     *             An SQL error occurs.
     */
    public void rebind(final CompositeName contextName, final String objectName, final Object object)
            throws NamingException {
        try {
            final String contextNameStr = contextName.toString();
            final Connection jdbcConnection = this.dataSource.getConnection();

            try {
                jdbcConnection.setAutoCommit(false);

                // 1 - We check if the bound object exist and update it in the
                // same time
                if (!this.updateBoundObject(contextNameStr, objectName, object, jdbcConnection)) {
                    // 2 - Else we create it
                    this.createBoundObject(contextNameStr, objectName, object, jdbcConnection);
                }

                jdbcConnection.commit();

            } finally {
                // We rollback potential updates. In case of successfully
                // processing a commit must occur before this point
                jdbcConnection.rollback();
                jdbcConnection.close();
            }
        } catch (final SQLException e) {
            throw new NamingException(e.getMessage());
        } catch (final IOException e) {
            throw new NamingException(e.getMessage());
        }
    }

    /**
     * Unbind a named object or delete a subcontext
     * 
     * @param contextName
     *            Context name, not null.
     * @param objectName
     *            Name of the object binding. If <code>null</code>, the
     *            subcontext <code>contextName</code> will be destroyed.
     * @throws NotContextException
     *             In case of context deletion, <code>contextName</code> is not
     *             a context.
     * @throws NameNotFoundException
     *             <code>contextName/objectName</code> does not exist.
     * @throws ContextNotEmptyException
     *             In case of context deletion, <code>contextName</code> is not
     *             empty.
     * @throws NamingException
     *             An SQL error occurs.
     */
    public void unbind(final CompositeName contextName, final String objectName)
            throws NotContextException, NameNotFoundException, ContextNotEmptyException,
            NamingException {

        try {
            final String contextNameStr = contextName.toString();
            final Connection jdbcConnection = this.dataSource.getConnection();

            try {
                jdbcConnection.setAutoCommit(false);

                if (objectName == null) {
                    // We must destroyed a sub-context:
                    // 1 - We check if it exist
                    final StringBuffer sbCheckContext = new StringBuffer(
                            "SELECT JNDI_CONTEXT.CONTEXT_ID, JNDI_BIND.NAME ").append(
                            "FROM JNDI_CONTEXT  LEFT JOIN JNDI_BIND ").append(
                            "ON JNDI_CONTEXT.CONTEXT_ID = JNDI_BIND.CONTEXT_ID ").append(
                            "WHERE JNDI_CONTEXT.CONTEXT = ? ");
                    // "WHERE JNDI_CONTEXT.CONTEXT = ? FOR UPDATE OF JNDI_CONTEXT");
                    final PreparedStatement pstCheckContext = jdbcConnection
                            .prepareStatement(sbCheckContext.toString());
                    int contextId;

                    try {
                        pstCheckContext.setString(1, contextNameStr);
                        final ResultSet rsCheckContext = pstCheckContext.executeQuery();
                        try {
                            if (rsCheckContext.next()) {
                                // The subcontext exists
                                if (rsCheckContext.getString(2) != null) {
                                    // Names exists under the context
                                    throw new ContextNotEmptyException(contextName
                                            + " is not empty.");
                                }
                                contextId = rsCheckContext.getInt(1);
                            } else {
                                throw new NameNotFoundException(contextName + " does not exist.");
                            }
                        } finally {
                            rsCheckContext.close();
                        }
                    } finally {
                        pstCheckContext.close();
                    }

                    // 2 - We remove it
                    this.destroyContextById(contextId, jdbcConnection);
                } else {
                    // We must destroyed a bound object:
                    // 1 - We check if it exist and remove it in the same time
                    if (!this.destroyBoundObject(contextNameStr, objectName, jdbcConnection)) {
                        throw new NameNotFoundException(contextName + " does not exist.");
                    }
                }

                jdbcConnection.commit();

            } finally {
                // We rollback potential updates. In case of successfully
                // processing a commit must occur before this point
                jdbcConnection.rollback();
                jdbcConnection.close();
            }
        } catch (final SQLException e) {
            throw new NamingException(e.getMessage());
        }
    }

    protected class IsContextException extends Exception {

        private static final long serialVersionUID = 4278892374765661555L;

        private final CompositeName contextName;

        public IsContextException(final CompositeName contextName) {
            this.contextName = contextName;
        }

        public CompositeName getContextName() {
            return contextName;
        }

    }

    /**
     * Retrieve a named object that can be a value object.or a context.
     * 
     * @param contextName
     *            Context name, not null.
     * @param objectName
     *            Name, only into <code>contextName</code>, no directory, of the
     *            named object to retrieve, not null.
     * @throws NameNotFoundException
     *             No named object found.
     * @throws NamingException
     *             An SQL error occurs.
     * @throws IsContextException
     *             The named object to retrieve is a context.
     * @return The named value object.
     */
    public Object lookup(final CompositeName contextName, final String objectName)
            throws IsContextException, NameNotFoundException, NamingException {
        try {
            final String contextNameStr = contextName.toString();
            final Connection jdbcConnection = this.dataSource.getConnection();

            try {
                jdbcConnection.setAutoCommit(false);

                // 1 - We get the bound object
                Object object = this.getBoundObject(contextNameStr, objectName, jdbcConnection);
                if (object == null) {
                    // The name object is not a value object, we check if it is
                    // a context
                    final CompositeName ctx = (CompositeName) contextName.clone();
                    ctx.add(objectName);
                    final String ctxStr = ctx.toString();
                    if (this.isContextExist(ctxStr, jdbcConnection)) {
                        throw new IsContextException(ctx);
                    }

                    // The bound object does not exist
                    throw new NameNotFoundException(ctxStr + " does not exist.");
                }

                jdbcConnection.commit();

                return object;

            } finally {
                // We rollback potential updates. In case of successfully
                // processing a commit must occur before this point
                jdbcConnection.rollback();
                jdbcConnection.close();
            }
        } catch (final SQLException e) {
            e.printStackTrace();
            throw new NamingException(e.getMessage());
        } catch (final IOException e) {
            throw new NamingException(e.getMessage());
        } catch (final ClassNotFoundException e) {
            throw new NamingException(e.getMessage());
        }
    }

    /**
     * Enumerates the names bound in a context with their object class name. In
     * case of a sub context, the object class name is set to <code>null</code>.
     * 
     * @param contextName
     *            Context name, not null.
     * @throws NameNotFoundException
     *             The context name does not exist.
     * @throws NamingException
     *             An SQL error occurs.
     * @return A map the names and class names of the bindings in this context.
     *         Each element of the map is of type
     *         <code>&lt;String, Object&gt;</code>, Object is <code>null</code>
     *         in case of a sub context.
     */
    public abstract Map<String, Object> list(final CompositeName contextName)
            throws NameNotFoundException, NamingException;

    private final static String isContextExistSql = new StringBuffer("SELECT CONTEXT_ID ").append(
            "FROM JNDI_CONTEXT ").append("WHERE CONTEXT = ? FOR UPDATE").toString();

    protected boolean isContextExist(final String contextName, final Connection jdbcConnection)
            throws SQLException {
        final PreparedStatement pstCheckContext = jdbcConnection
                .prepareStatement(isContextExistSql);

        try {
            pstCheckContext.setString(1, contextName);
            final ResultSet rsCheckContext = pstCheckContext.executeQuery();
            try {
                if (rsCheckContext.next()) {
                    // The subcontext exists
                    return true;
                } else {
                    return false;
                }
            } finally {
                rsCheckContext.close();
            }
        } finally {
            pstCheckContext.close();
        }
    }

    private final static String createContextSql = new StringBuffer(
            "INSERT INTO JNDI_CONTEXT (CONTEXT) VALUES (?)").toString();

    private void createContext(final String contextName, final Connection jdbcConnection)
            throws SQLException {
        final PreparedStatement pstCreateContext = jdbcConnection
                .prepareStatement(createContextSql);

        try {
            pstCreateContext.setString(1, contextName);
            pstCreateContext.executeUpdate();
        } finally {
            pstCreateContext.close();
        }
    }

    private final static String destroyContextByIdSql = new StringBuffer(
            "DELETE FROM JNDI_CONTEXT WHERE CONTEXT_ID = ?").toString();

    private void destroyContextById(final int contextId, final Connection jdbcConnection)
            throws SQLException {
        final PreparedStatement pstDestroyContext = jdbcConnection
                .prepareStatement(destroyContextByIdSql);

        try {
            pstDestroyContext.setInt(1, contextId);
            pstDestroyContext.executeUpdate();
        } finally {
            pstDestroyContext.close();
        }
    }

    private final static String isBoundObjectExistWithoutLockSql = new StringBuffer(
            "SELECT JNDI_CONTEXT.CONTEXT_ID, JNDI_BIND.NAME ").append(
            "FROM JNDI_CONTEXT LEFT JOIN JNDI_BIND ").append(
            "ON JNDI_CONTEXT.CONTEXT_ID = JNDI_BIND.CONTEXT_ID ").append(
            "WHERE JNDI_CONTEXT.CONTEXT = ? ").append("AND JNDI_BIND.NAME = ?").toString();

    private boolean isBoundObjectExistWithoutLock(final String contextName,
            final String objectName, final Connection jdbcConnection) throws SQLException {

        final PreparedStatement pstCheckContext = jdbcConnection
                .prepareStatement(isBoundObjectExistWithoutLockSql);

        try {
            pstCheckContext.setString(1, contextName);
            pstCheckContext.setString(2, objectName);
            final ResultSet rsCheckContext = pstCheckContext.executeQuery();
            try {
                if (rsCheckContext.next()) {
                    // The bound object exists
                    return true;
                } else {
                    return false;
                }
            } finally {
                rsCheckContext.close();
            }
        } finally {
            pstCheckContext.close();
        }
    }

    private final static String createBoundObjectSql = new StringBuffer("INSERT INTO JNDI_BIND ")
            .append("SELECT CONTEXT_ID, ?, ? FROM JNDI_CONTEXT WHERE CONTEXT = ?").toString();

    private void createBoundObject(final String contextName, final String objectName,
            final Object object, final Connection jdbcConnection) throws SQLException, IOException {
        final PreparedStatement pstCreateBoundObject = jdbcConnection
                .prepareStatement(createBoundObjectSql);

        try {

            pstCreateBoundObject.setString(1, objectName);
            RegistryClient.writeObject(pstCreateBoundObject, 2, object);
            pstCreateBoundObject.setString(3, contextName);
            pstCreateBoundObject.executeUpdate();

        } finally {
            pstCreateBoundObject.close();
        }
    }

    private static final void writeObject(final PreparedStatement pst, final int idx,
            final Object object) throws SQLException, IOException {
        final ByteArrayOutputStream baos = new ByteArrayOutputStream();
        final ObjectOutputStream oos = new ObjectOutputStream(baos);
        oos.writeObject(object);
        baos.close();

        pst.setBinaryStream(idx, new ByteArrayInputStream(baos.toByteArray()), baos.size());
    }

    private final static String getBoundObjectSql = new StringBuffer("SELECT JNDI_BIND.OBJECT ")
            .append("FROM JNDI_CONTEXT INNER JOIN JNDI_BIND ").append(
                    "ON JNDI_CONTEXT.CONTEXT_ID = JNDI_BIND.CONTEXT_ID ").append(
                    "WHERE JNDI_CONTEXT.CONTEXT = ? ").append("AND JNDI_BIND.NAME = ?").toString();

    private Object getBoundObject(final String contextName, final String objectName,
            final Connection jdbcConnection) throws SQLException, IOException,
            ClassNotFoundException {
        final PreparedStatement pstGetBoundObject = jdbcConnection
                .prepareStatement(getBoundObjectSql);

        try {

            pstGetBoundObject.setString(1, contextName);
            pstGetBoundObject.setString(2, objectName);
            final ResultSet rs = pstGetBoundObject.executeQuery();
            try {
                if (rs.next()) {
                    return RegistryClient.readObject(rs, 1);

                } else {
                    return null;
                }
            } finally {
                rs.close();
            }

        } finally {
            pstGetBoundObject.close();
        }
    }

    protected static final Object readObject(final ResultSet rs, final int idx)
            throws SQLException, IOException, ClassNotFoundException {

        final InputStream is = rs.getBinaryStream(idx);
        final ObjectInputStream ois = new ObjectInputStream(is);
        return ois.readObject();
    }

    private final static String updateBoundObjectSql = new StringBuffer("UPDATE JNDI_BIND ")
            .append("SET OBJECT = ? ").append(
                    "WHERE CONTEXT_ID = (SELECT CONTEXT_ID FROM JNDI_CONTEXT WHERE CONTEXT = ?) ")
            .append("AND NAME = ?").toString();

    /**
     * @param contextName
     * @param objectName
     * @param object
     * @param jdbcConnection
     * @return false if no object has been updated, true otherwise
     * @throws SQLException
     * @throws IOException
     */
    private boolean updateBoundObject(final String contextName, final String objectName,
            final Object object, final Connection jdbcConnection) throws SQLException, IOException {
        final PreparedStatement pstCreateBoundObject = jdbcConnection
                .prepareStatement(updateBoundObjectSql);

        try {

            RegistryClient.writeObject(pstCreateBoundObject, 1, object);
            pstCreateBoundObject.setString(2, contextName);
            pstCreateBoundObject.setString(3, objectName);
            return pstCreateBoundObject.executeUpdate() != 0;

        } finally {
            pstCreateBoundObject.close();
        }
    }

    private final static String destroyBoundObjectSql = new StringBuffer("DELETE FROM JNDI_BIND ")
            .append("WHERE NAME = ? ").append(
                    "AND CONTEXT_ID = (SELECT CONTEXT_ID FROM JNDI_CONTEXT WHERE CONTEXT = ? )")
            .toString();

    /**
     * @param contextName
     * @param objectName
     * @param jdbcConnection
     * @return false if no object has been destroyed, true otherwise
     * @throws SQLException
     */
    private boolean destroyBoundObject(final String contextName, final String objectName,
            final Connection jdbcConnection) throws SQLException {
        final PreparedStatement pstDestroyBoundObject = jdbcConnection
                .prepareStatement(destroyBoundObjectSql);

        try {
            pstDestroyBoundObject.setString(1, objectName);
            pstDestroyBoundObject.setString(2, contextName);
            return pstDestroyBoundObject.executeUpdate() != 0;
        } finally {
            pstDestroyBoundObject.close();
        }
    }

}
