package org.ow2.petals.registry.database;

import java.util.Hashtable;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import javax.naming.Binding;
import javax.naming.CompositeName;
import javax.naming.Context;
import javax.naming.InvalidNameException;
import javax.naming.LinkRef;
import javax.naming.Name;
import javax.naming.NameClassPair;
import javax.naming.NameParser;
import javax.naming.NamingEnumeration;
import javax.naming.NamingException;
import javax.naming.Reference;
import javax.naming.Referenceable;

import org.ow2.petals.registry.database.RegistryClient.IsContextException;
import org.ow2.petals.util.LoggingUtil;

public class RegistryContextImpl implements Context {

    private final String connectionUrl;

    private final String username;

    private final String password;

    private final int connectionPoolSize;

    private final int batchsize;

    private final RegistryClient postgreSqlClient;

    final private CompositeName contextPath;

    /**
     * Logging utilities.
     */
    final private LoggingUtil log;

    /**
     * SGBD names
     */
    private enum SGBDNames {
        mysql, postgresql
    };

    /**
     * Constructor to used to create the initial context.
     * 
     * @param connectionUrl
     * @param username
     * @param password
     * @param connectionPoolSize
     * @param batchsize
     * @param log
     * @throws NamingException
     */
    public RegistryContextImpl(final String connectionUrl, final String username,
            final String password, final int connectionPoolSize, final int batchsize,
            final LoggingUtil log) throws NamingException {

        this(connectionUrl, username, password, connectionPoolSize, batchsize, new CompositeName(
                RegistryClient.ROOT_CONTEXT), true, log);

    }

    public RegistryContextImpl(final String connectionUrl, final String username,
            final String password, final int connectionPoolSize, final int batchsize,
            final CompositeName contextPath, final boolean checkStructures, final LoggingUtil log)
            throws NamingException {

        this.connectionUrl = connectionUrl;
        this.username = username;
        this.password = password;
        this.connectionPoolSize = connectionPoolSize;
        this.batchsize = batchsize;
        this.contextPath = contextPath;
        this.log = log;

        if (connectionUrl.contains(SGBDNames.mysql.toString()))
            this.postgreSqlClient = new MySqlClient(connectionUrl, username, password,
                    connectionPoolSize, batchsize, checkStructures, log);
        else if (connectionUrl.contains(SGBDNames.postgresql.toString()))
            this.postgreSqlClient = new PostgreSqlClient(connectionUrl, username, password,
                    connectionPoolSize, batchsize, checkStructures, log);
        else
            throw new NamingException(" Unsupported SGBD ");

    }

    public void bind(Name name, Object obj) throws NamingException {
        bind(name.toString(), obj);
    }

    public void bind(String name, Object obj) throws NamingException {
        this.log.start('(' + name + ',' + obj + ')');

        try {
            this.log.debug("Class name of the written object: " + obj.getClass().getName());
            if (obj instanceof Referenceable) {
                obj = ((Referenceable) obj).getReference();
            }

            this.postgreSqlClient.bind(this.contextPath, name, obj);
        } finally {
            this.log.end();
        }

    }

    public void rebind(Name name, Object obj) throws NamingException {
        rebind(name.toString(), obj);
    }

    public void rebind(String name, Object obj) throws NamingException {
        this.log.start('(' + name + ',' + obj + ')');

        try {
            this.log.debug("Class name of the written object: " + obj.getClass().getName());
            if (obj instanceof Referenceable) {
                obj = ((Referenceable) obj).getReference();
            }

            this.postgreSqlClient.rebind(this.contextPath, name, obj);
        } finally {
            this.log.end();
        }
    }

    public Object lookup(Name name) throws NamingException {
        return lookup(name.toString());
    }

    public Object lookup(String name) throws NamingException {
        this.log.start('(' + name + ')');

        try {

            if (name == null || name.length() == 0) {
                return new RegistryContextImpl(this.connectionUrl, this.username, this.password,
                        this.connectionPoolSize, this.batchsize, this.contextPath, false, this.log);
            } else {
                Object object;
                try {
                    final CompositeName fullName = new CompositeName(name);
                    fullName.addAll(0, this.contextPath);

                    object = this.postgreSqlClient.lookup((CompositeName) fullName
                            .getPrefix(fullName.size() - 1), fullName
                            .getSuffix(fullName.size() - 1).toString());
                } catch (final IsContextException e) {
                    // The name is not a value object but a context
                    object = new RegistryContextImpl(this.connectionUrl, this.username,
                            this.password, this.connectionPoolSize, this.batchsize, e
                                    .getContextName(), false, this.log);
                }
                this.log.debug("Class name of the read object: " + object.getClass().getName());

                // If the object is a LinkRef, resolve the link
                if (object instanceof LinkRef) {
                    String linkedName = ((LinkRef) object).getLinkName();
                    if (linkedName.startsWith(".")) {
                        // relative to the current context
                        object = this.lookup(linkedName.substring(1));
                    } else {
                        // relative to the root context
                        object = new RegistryContextImpl(this.connectionUrl, this.username,
                                this.password, this.connectionPoolSize, this.batchsize, this.log)
                                .lookup(linkedName);
                    }
                }

                object = RegistryContextImpl.resolveObject(object);
                return object;
            }

        } finally {
            this.log.end();
        }
    }

    public void unbind(Name name) throws NamingException {
        unbind(name.toString());
    }

    public void unbind(String name) throws NamingException {
        this.log.start('(' + name + ')');

        try {
            this.postgreSqlClient.unbind(this.contextPath, name);
        } finally {
            this.log.end();
        }
    }

    public void close() throws NamingException {
        this.postgreSqlClient.close();
    }

    public Hashtable<?, ?> getEnvironment() throws NamingException {
        throw (new NamingException("Not yet available"));
    }

    public NamingEnumeration<NameClassPair> list(Name name) throws NamingException {
        return list(name.toString());
    }

    public NamingEnumeration<NameClassPair> list(String name) throws NamingException {
        this.log.start('(' + name + ')');
        try {

            final Map<String, Object> liste = this.postgreSqlClient.list(RegistryContextImpl
                    .concat(this.contextPath, name));
            this.log.debug(liste.size() + " objects returned.");
            final List<NameClassPair> enumeration = new LinkedList<NameClassPair>();
            for (final Entry<String, Object> entry : liste.entrySet()) {
                final Object object = entry.getValue();
                if (object == null) {
                    // Its a subcontext
                    enumeration.add(new NameClassPair(entry.getKey(), Context.class.getName()));
                } else {
                    // Its a named object
                    enumeration.add(new NameClassPair(entry.getKey(), object.getClass().getName()));
                }
            }
            return new NamingEnumerationImpl<NameClassPair>(enumeration);
        } finally {
            this.log.end();
        }
    }

    public NamingEnumeration<Binding> listBindings(Name name) throws NamingException {
        return listBindings(name.toString());
    }

    public NamingEnumeration<Binding> listBindings(String name) throws NamingException {

        this.log.start('(' + name + ')');
        try {

            final Map<String, Object> liste = this.postgreSqlClient.list(RegistryContextImpl
                    .concat(this.contextPath, name));
            this.log.debug(liste.size() + " object(s) returned.");
            final List<Binding> bindings = new LinkedList<Binding>();
            for (Entry<String, Object> entry : liste.entrySet()) {

                Object object = entry.getValue();
                if (object == null) {
                    // 1- resolve context and add it
                    final CompositeName newCtxPath = (CompositeName) RegistryContextImpl.concat(
                            contextPath, name).clone();
                    bindings.add(new Binding(entry.getKey(), new RegistryContextImpl(
                            this.connectionUrl, this.username, this.password,
                            this.connectionPoolSize, this.batchsize, newCtxPath, false, this.log)));
                } else {
                    // If the object is a LinkRef, resolve the link
                    if (object instanceof LinkRef) {
                        String linkedName = ((LinkRef) object).getLinkName();
                        if (linkedName.startsWith(".")) {
                            // relative to the current context
                            object = this.lookup(linkedName.substring(1));
                        } else {
                            // relative to the root context
                            object = new RegistryContextImpl(this.connectionUrl, this.username,
                                    this.password, this.connectionPoolSize, this.batchsize,
                                    this.log).lookup(linkedName);
                        }
                    }

                    // 2- Add object resolving references
                    bindings.add(new Binding(entry.getKey(), RegistryContextImpl
                            .resolveObject(object)));
                }
            }

            return new NamingEnumerationImpl<Binding>(bindings);
        } finally {
            this.log.end();
        }
    }

    public Context createSubcontext(Name name) throws NamingException {
        return this.createSubcontext(name.toString());
    }

    public Context createSubcontext(String name) throws NamingException {
        this.log.start('(' + name + ')');

        try {
            final CompositeName subPath = RegistryContextImpl.concat(this.contextPath, name);
            this.postgreSqlClient.bind(subPath, null, null);

            return new RegistryContextImpl(this.connectionUrl, this.username, this.password,
                    this.connectionPoolSize, this.batchsize, subPath, false, this.log);
        } finally {
            this.log.end();
        }
    }

    public void destroySubcontext(Name name) throws NamingException {
        this.destroySubcontext(name.toString());
    }

    public void destroySubcontext(String name) throws NamingException {
        this.log.start('(' + name + ')');

        try {
            final CompositeName subPath = RegistryContextImpl.concat(this.contextPath, name);
            this.postgreSqlClient.unbind(subPath, null);

        } finally {
            this.log.end();
        }
    }

    public String getNameInNamespace() throws NamingException {
        return this.contextPath.toString();
    }

    public Object lookupLink(Name name) throws NamingException {
        return this.lookup(name);
    }

    public Object lookupLink(String name) throws NamingException {
        return this.lookup(name);
    }

    public Object addToEnvironment(Name propName, Object propVal) throws NamingException {
        throw (new NamingException("Not yet available"));
    }

    public Object addToEnvironment(String propName, Object propVal) throws NamingException {
        throw (new NamingException("Not yet available"));
    }

    public Name composeName(Name name, Name prefix) throws NamingException {
        throw (new NamingException("Not yet available"));
    }

    public String composeName(String name, String prefix) throws NamingException {
        throw (new NamingException("Not yet available"));
    }

    public NameParser getNameParser(Name name) throws NamingException {
        throw (new NamingException("Not yet available"));
    }

    public NameParser getNameParser(String name) throws NamingException {
        throw (new NamingException("Not yet available"));
    }

    public Object removeFromEnvironment(String propName) throws NamingException {
        throw (new NamingException("Not yet available"));
    }

    public void rename(Name oldName, Name newName) throws NamingException {
        throw (new NamingException("Not yet available"));
    }

    public void rename(String oldName, String newName) throws NamingException {
        throw (new NamingException("Not yet available"));
    }

    /**
     * Add a name in a composite name at last position.
     * 
     * @param compositeName
     *            The composite name at which a name must be added in last
     *            position. Not null.
     * @param name
     *            The name to add to <code>compositeName</code> in last. Not
     *            null.
     * @return A new composite name as result of a concatenation of
     *         <code>compositeName</code> and <code>name</code>.
     * @throws NamingException
     */
    private static CompositeName concat(CompositeName compositeName, String name)
            throws NamingException {
        if (name == null)
            throw new InvalidNameException();
        /*
         * CompositeName res = new CompositeName(name); //trim(res); for (int i
         * = path.size() - 1; i > -1 ; i--) { res.add(0, path.get(i)); } return
         * res;
         */

        final CompositeName res = (CompositeName) compositeName.clone();
        if (name.length() > 0 && !"/".equals(name)) {
            res.add(name);
        }
        return res;

    }

    public final static Object resolveObject(final Object obj) throws NamingException {
        if (obj instanceof Reference) {
            try {
                return javax.naming.spi.NamingManager.getObjectInstance(obj, null, null, null);
            } catch (final Exception e) {
                throw new NamingException(e.getMessage());
            }
        } else {
            return obj;
        }

    }

    /*
     * private static CompositeName trim(CompositeName name) throws
     * NamingException { int i = 0; while (i < name.size()) { String s =
     * (String)name.remove(i); s = s.trim(); if (s.length() > 0) { name.add(i,
     * s); i++; } } return name; }
     */
}
